#!/usr/bin/env python
# -*- coding: utf-8 -*-
# simulation.py
# © Claus Albowski

# Paket fuer mathematische Funktionen importieren
from math import sqrt
def wurzel(x): return sqrt(x)

import term
from math import *
from funktionen import *

#### globals() gibt das globale Dictionary
#### von Paaren aus Name und Objekt zurueck
import platform
versionsNummer=platform.python_version()[0]

#### Oberklasse für alle Faktoren ====================================
class Faktor(object):
    '''Klasse für Faktoren'''

    # Klassenvariable fuer alle Faktornamen
    alleFaktorNamen=[]

    @staticmethod
    def GibAlleFaktorNamen():
        return Faktor.alleFaktorNamen

    @staticmethod
    def FaktorZuNamen(name):
        '''gibt aus der Liste der Faktoren zum Namen das Objekt zurück;
        im Misserfolgsfall None'''
        if name in globals().keys():
            return globals()[name]
        return None

    @staticmethod
    def LoescheFaktor(faktorName):
        '''loescht sowohl das Objekt aus dem globalen Namensraum
        als auch den Namen aus der Liste alleFaktoren'''
        globals().pop(faktorName)
        Faktor.alleFaktorNamen.remove(faktorName)
    
    #### Konstruktor von Faktor
    def __init__(self, name, startwert=0.0):
        '''intern werden die Namen der Objekte gespeichert und
        im globalen Namensraum ein Objekt mit dem Namen angelegt'''
        while name in globals().keys():
            name=input(name+'existiert bereits! aendern in: ')
        if self.PruefeNamen(name):
            self.__name=name
            Faktor.alleFaktorNamen.append(name)
        self.SetzeStartwert(startwert) # setzt auch den Wert
        globals().update({name:self})
        # In dieser Version werden alle alten Werte gespeichert
        self.__alleWerte = [startwert]

    def __call__(self, verzoegerung=0):
        '''ermoeglicht (in dieser Form allein) den Abruf
        des aktuellen oder eines alten Wertes durch Aufruf des Objektnamen'''
        try: verzoegerung=int(verzoegerung)
        except ValueError as e:
            self.Meldung(str(e))
            verzoegerung=0
        ## welcher Wert zurueckgegeben werden muss, haengt von einer
        ## moeglichen Verzoegerung ab
        if (verzoegerung==0): return self.GibWert()
        index=max(0,len(self.__alleWerte)-1 - verzoegerung + int(dt))
        ## Index zulaessig?
        index=min(len(self.__alleWerte)-1, index)
        return self.__alleWerte[index]

    def GibAlleWerte(self):
        '''liefert die Liste aller bisherigen Werte zurueck'''
        return self.__alleWerte

    def UpdateAlleWerte(self):
        '''fuegt den aktuellen Wert hinzu'''
        self.__alleWerte.append(self.GibWert())

    def GibNamen(self):
        try:
            return self.__name
        except AttributeError as e:
            self.Meldung("Name ist nicht richtig initialisiert:\n"
                         +"SetzeNamen() mit String notwendig!")
    
    def PruefeNamen(self, name):
        try:
            assert name.__class__== str,\
                   "Der Name muss den Typ String haben."
            return True
        except AssertionErro as e:
            self.Meldung(str(e))
            return False
    
    def SetzeNeuenNamen(self, name):
        if not(self.PruefeNamen(name)): return
        while name in globals().keys():
            name=input(name+u' existiert bereits!\nÄndern oder Abbruch mit leerer Eingabe.')
            if name=='': return
        ## Objekt mit dem bisherigen Namen muss im globalen
        ## Namensraum geloescht werden
        Faktor.LoescheFaktor(self.__name)
        self.__name=name
        Faktor.alleFaktorNamen.append(name)
        globals().update({name:self})

    def SetzeStartwert(self, startwert=0.0):
        '''setzt den Startwert und initialisiert den Wert'''
        if startwert.__class__==int:
            startwert=float(startwert)
        try:
            assert startwert.__class__== float,\
                   "Der Parameter muss den Typ float haben."
            self.__startwert=startwert
            self.__wert=startwert
        except AssertionError as e:
            self.Meldung(str(e))
        self.Reset()

    def Reset(self):
        '''Setzt den Wert auf den Startwert zurueck.'''
        self.__wert=self.GibStartwert()
        self.__alleWerte = [self.GibStartwert()]

    def GibStartwert(self):
        try:
            return self.__startwert
        except AttributeError as e:
            self.Meldung("Startwert ist nicht richtig initialisiert:\n"
                         +"oder Objekt hat keinen Startwert.\n"
                         +"SetzeStartwert(...) mit float notwendig!")

    def GibWert(self):
        '''gibt den aktuellen Wert zurück'''
        try:
            return self.__wert
        except AttributeError as e:
            self.Meldung("Wert ist nicht richtig initialisiert:\n"
                         +"SetzeStartwert(...) mit float notwendig!")
    
    def Meldung(self, text):
        Simulation.GibSimulation().Meldung(text)

#### Klasse fuer Bestandsfaktoren --------------------------------
class Bestandsfaktor(Faktor):
    '''Klasse für Bestandsfaktoren'''

    #### Konstruktor von Bestandsfaktor
    def __init__(self, name, startwert=0.0):
        Faktor.__init__(self,name,startwert)
        self.__zuFluesse=[] # zu diesem Faktor!
        self.__abFluesse=[] # von diesem Faktor!

    def BerechneNeuenWert(self, halberSchritt=False):
        if halberSchritt: delta=0.5*dt
        else: delta=dt
        self.__neuerWert=self._Faktor__wert
        for fluss in self.__zuFluesse:
            self.__neuerWert += eval(fluss).GibWert()*delta
        for fluss in self.__abFluesse:
            self.__neuerWert -= eval(fluss).GibWert()*delta

    def UebernimmNeuenWert(self, neuerWert=None):
        '''die Methode ist notwendig fuer die Iteration;
        sie darf aber nur von Bestandsfaktoren verwendet werden.'''
        if neuerWert==None: neuerWert = self.__neuerWert
        self._Faktor__wert = neuerWert

    def GibZuFluesse(self):
        return self.__zuFluesse

    def GibAbFluesse(self):
        return self.__abFluesse
    
    def AendereFlussNamen(self, alterName,neuerName):
        for fluss in self.__zuFluesse:
            if fluss==alterName:
                self.__zuFluesse.remove(fluss)
                self.__zuFluesse.append(neuerName)
        for fluss in self.__abFluesse:
            if fluss==alterName:
                self.__abFluesse.remove(fluss)
                self.__abFluesse.append(neuerName)

    def Neuer_Fluss(self, flussName, verstaerkend=True):
        if verstaerkend:
            self.__zuFluesse.append(flussName)
        else:
            self.__abFluesse.append(flussName)

    def Entferne_Fluss(self, flussName):
        if flussName in self.__zuFluesse:
            self.__zuFluesse.remove(flussName)
        if flussName in self.__abFluesse:
            self.__abFluesse.remove(flussName)

#### Klasse fuer Fluesse ---------------------------------------
class Fluss(Faktor):
    '''Klasse fuer Fluesse'''
    
    #### Konstruktor von Fluss
    def __init__(self, name, aus_Bestand='Quelle', in_Bestand='Senke'):
        '''Der Fluss muss so initialisiert werden, dass er mindestens einen
        Ausgangs-Bestand oder einen Ziel-Bestand hat.'''
        assert (aus_Bestand!='Quelle') or (in_Bestand!='Senke'), "Ausgang oder Ziel notwendig!"
        Faktor.__init__(self,name)
        self.__aus=aus_Bestand
        if aus_Bestand!='Quelle':
            eval(aus_Bestand).Neuer_Fluss(name, False)
        self.__in=in_Bestand
        if in_Bestand!='Senke':
            eval(in_Bestand).Neuer_Fluss(name, True)
        self.__term="0.0" # nur dummy
        self.SetzeStartwert(0.0) # nur dummy
        self.__wirkFaktorNamen=[]

    def GibAusgangsBestand(self):
        return self.__aus

    def GibZielBestand(self):
        return self.__in

    def AendereAusgangsBestand(self, alterName, neuerName):
        if self.__aus==alterName: self.__aus=neuerName

    def AendereZielBestand(self, alterName, neuerName):
        if self.__in==alterName: self.__in=neuerName

    def GibWert(self):
        '''überschreibt die GibWert-Methode von Faktor'''
        return eval(self.__term)

    def SetzeTerm(self,term):
        for name in self.__wirkFaktorNamen:
            if not(name+"(" in term):
                self.Meldung("Faktor "+name+"(...) nicht enthalten")
                return False
        self.__term=term
        return True

    def UpdateTerm(self, alterName, neuerName):
        '''Passt den Term ggf. an einen geaenderten Faktornamen an.'''
        if alterName in self.__wirkFaktorNamen:
            self.__wirkFaktorNamen.remove(alterName)
            self.__wirkFaktorNamen.append(neuerName)
            self.SetzeTerm(term.UpdateTerm(self.__term, alterName, neuerName))
    
    def GibTerm(self):
        return self.__term
    
    def NeuerWirkFaktor(self, faktorName):
        '''fuegt einen einwirkenden Faktor hinzu,
        ändert nicht Ausgangs-Bestand oder Ziel-Bestand'''
        if faktorName in self.__wirkFaktorNamen: self.Meldung("Faktor bereits enthalten!")
        else: self.__wirkFaktorNamen.append(faktorName)

    def EntferneWirkFaktor(self, faktorName):
        '''entfernt einen einwirkenden Faktor,
        ändert nicht Ausgangs-Bestand oder Ziel-Bestand'''
        if faktorName in self.__wirkFaktorNamen:
            self.__wirkFaktorNamen.remove(faktorName)

    def ZeigeFaktorNamen(self):
        '''gibt alle einwirkenden Faktoren zurück,
        nicht Ausgangs-Bestand oder Ziel-Bestand,
        es sei denn, sie haben eine Wirkung auf den Fluss'''
        return self.__wirkFaktorNamen 

#### Klasse für Faktoren, die einen Parameter darstellen ----------
class Parameter(Faktor):
    '''Klasse für Faktoren ohne Einwirkungen
    Tabellenwerte und Terme sind moeglich'''
    
    #### Konstruktor von Parameter
    def __init__(self, name, startwert=0.0):
        '''ImKonstruktor keinen Term!'''
        self.__tabelle = None
        self.__term = None
        Faktor.__init__(self,name,startwert)

    def GibWert(self):
        '''gibt den aktuellen Wert zurück'''
        if self.__tabelle!=None:
            return self.__tabelle.GibWert(t())
        elif self.__term!=None:
            return eval(self.__term)
        return self.GibStartwert()

    def SetzeStartwert(self, wert=0.0):
        '''setzt den Startwert und den aktuellen Wert und
        deaktiviert die Tabellenfunktion und den Term'''
        super(Parameter,self).SetzeStartwert(wert)
        self.__tabelle = None
        self.__term = None

    def GibStartWert(self):
        if self.__tabelle!=None:
            return self.__tabelle.GibWert(0.0)
        return super(Parameter,self).GibStartwert()

    def GibTabelle(self):
        return self.__tabelle
    
    def GibTabellenWerte(self):
        return self.__tabelle.GibTabelle()

    def SetzeTabelle(self, zeiten, werte):
        self.__tabelle = Tabelle(zeiten,werte)
        self.Reset()
    
    def NeuesTabellenPaar(self, wertePaar):
        """Erzeugt eine neue Tabelle, wenn noch keine existiert."""
        if self.__tabelle==None:
            self.Reset()
            self.__tabelle = Tabelle([0.0],[self.GibStartwert()])
        if wertePaar[0]==0.0:
            super(Parameter,self).SetzeStartwert(wertePaar[1])
        self.__tabelle.NeuesWertePaar(wertePaar)

    def GibTerm(self):
        return self.__term

    def SetzeTerm(self,term):
        for name in Faktor.GibAlleFaktorNamen():
            if (name!='t') and (name+"(" in term):
                self.Meldung(u"Parameter enthalten keine Wirkfaktoren\n"+
                             u"Zwischengrößen verwenden!")
                return False
        self.__term=term
        return True
        
#### Klasse für Faktoren, die Zwischengroessen darstellen ---------
class Zwischengroesse(Faktor):
    '''Klasse für Zwischengrößen
    Einwirkende Faktoren sind nur nach ihren Namen gespeichert.'''
    
    @staticmethod
    def EntferneZwischengroesse(name):
        '''entfernt das Zwischengroessenobjekt zum Namen'''
        faktor=Faktor.FaktorZuNamen(name)
        if name!=None:
            ## was ist speziell bei Zwischengroessen zu tun?
            pass
        Faktoren.LoescheFaktor(faktor)
    
    #### Konstruktor von Zwischengroesse
    def __init__(self, name):
        '''Zwischengrößen vereinfachen die Beschreibung.'''
        Faktor.__init__(self,name)
        self.__term="0.0" # nur dummy
        self.SetzeStartwert(0.0) # nur dummy
        self.__wirkFaktorNamen=[]

    def GibWert(self):
        '''überschreibt die GibWert-Methode von Faktor'''
        return eval(self.__term)

    def SetzeTerm(self,term):
        for name in self.__wirkFaktorNamen:
            if not(name+"()" in term):
                self.Meldung("Faktor "+name+" nicht enthalten")
                return False
        self.__term=term
        return True

    def UpdateTerm(self, alterName, neuerName):
        '''Passt den Term ggf. an einen geaenderten Faktornamen an.'''
        if alterName in self.__wirkFaktorNamen:
            self.__wirkFaktorNamen.remove(alterName)
            self.__wirkFaktorNamen.append(neuerName)
            self.SetzeTerm(term.UpdateTerm(self.__term, alterName, neuerName))
    
    def GibTerm(self):
        return self.__term
    
    def NeuerWirkFaktor(self, faktorName):
        '''fuegt einen einwirkenden Faktor hinzu'''
        if faktorName in self.__wirkFaktorNamen: self.Meldung("Faktor bereits enthalten!")
        else: self.__wirkFaktorNamen.append(faktorName)

    def EntferneWirkFaktor(self, faktorName):
        '''entfernt einen einwirkenden Faktor'''
        if faktorName in self.__wirkFaktorNamen:
            self.__wirkFaktorNamen.remove(faktorName)

    def ZeigeFaktorNamen(self):
        return self.__wirkFaktorNamen

#### Klasse fuer Wirkungen -------------------------------------------
class Wirkung():
    '''Klasse für Wirkungen
    verwaltet werden die Namen'''
    wirkungen=[]

    @staticmethod
    def HatWirkung(vonName, aufName):
        '''Gibt die Wirkung zurueck, falls sie existiert, sonst False'''
        for wirkung in Wirkung.wirkungen:
            if ((wirkung.GibVon()==vonName)
                and(wirkung.GibAuf()==aufName)): return wirkung
        return False
    
    @staticmethod
    def GibAlleWirkungen():
        return Wirkung.wirkungen
    
    @staticmethod
    def LoescheWirkung(wirkung):
        von,auf=wirkung.GibVon(),wirkung.GibAuf()
        faktor=Faktor.FaktorZuNamen(auf)
        faktor.EntferneWirkFaktor(von)
        Wirkung.wirkungen.remove(wirkung)

    @staticmethod
    def LoescheAlleWirkungen():
        wirkungen=[]
    
    @staticmethod
    def LoescheWirkungenZuFaktor(faktorName):
        '''Faktoren, die entfernt werden sollen, dürfen keine Wirkungen haben.'''
        for wirkung in Wirkung.wirkungen:
            if wirkung.Ausgang()==faktorName:
                Wirkung.wirkungen.remove(wirkung)
            if wirkung.Ziel()==faktorName:
                faktor.EntferneWirkFaktor(wirkung.Ausgang())
                Wirkung.wirkungen.remove(wirkung)

    #### Konstruktor fuer Wirkungen ------------------------------------    
    def __init__(self, vonName, aufName):
        '''Eine Wirkung muss so initialisiert werden, dass sowohl der
        Ausgangs-Faktor als auch der Ziel-Faktor gegeben sind.
        Keine Wirkung auf Bestandsfaktoren! Das sind Fluesse.
        Keine Wirkung auf Parameter! Das sind Zwischengroessen.'''
        geht = True
        self.__von=vonName
        von=Faktor.FaktorZuNamen(vonName)
        self.__auf=aufName
        auf=Faktor.FaktorZuNamen(aufName)
        if Wirkung.HatWirkung(vonName, aufName):
            self.Meldung('Wirkung existiert bereits')
            geht=False
        if von==None:
            self.Meldung(vonName+' existiert nicht')
            geht=False
        if auf==None:
            self.Meldung(aufName+' existiert nicht')
            geht=False
        if geht and (auf.__class__==Bestandsfaktor):
            self.Meldung('Keine Wirkungen auf Bestandsfaktoren!')
            geht=False
        if geht and (auf.__class__==Parameter):
            self.Meldung('Keine Wirkungen auf Parameter!')
            geht=False
        if geht:
            Wirkung.wirkungen.append(self)
            auf.NeuerWirkFaktor(vonName)
        else:
            self.Meldung(vonName+'->'+aufName+' geht nicht!')
        self.undefiniert=not(geht)

    def GibVon(self):
        return self.__von

    def GibAuf(self):
        return self.__auf

    def Meldung(self, text):
        Simulation.GibSimulation().Meldung(text)

    def FaktorNameGeaendert(self, alterName,neuerName):
        if self.GibVon()==alterName:
            self.__von=neuerName
        if self.GibAuf()==alterName:
            self.__auf=neuerName

#### Hilfsobjekt fuer die Zeit -------------------------------------------------
class Zeit(Faktor):
    '''Erzeugt ein t-Objekt als Singleton'''

    dasZeitObjekt=None
    @staticmethod
    def GibZeit():
        if Zeit.dasZeitObjekt==None:
            Zeit.dasZeitObjekt=Zeit()
        return Zeit.dasZeitObjekt
    
    def __init__(self):
        '''Der Startwert sollte 0.0 sein.'''
        assert Zeit.dasZeitObjekt==None, "Konstruktor nicht direkt aufrufen"
        Faktor.__init__(self, 't', 0.0)        
        self.Setze_dt()
        # es gibt kein Attribut self.__dt, nur eine globale Variable dt

    def Setze_dt(self, dt=1.0):
        '''Definiert neben dem Attribut
        auch die globale Variable mit dem uebergebenen Wert'''
        if type(dt) is int: dt=float(dt)
        if not (type(dt) is float):
            self.Meldung('Zahl bei dt eingeben')
            return
        if dt<=0:
            self.Meldung('positive Zahl bei dt eingeben')
            return
        wertZuweisung='dt='+str(dt)
        exec(wertZuweisung,globals())

    def Gib_dt(self):
        return eval('dt')

    def Schritt(self):
        '''Setzt den wert von t um dt weiter'''
        self._Faktor__wert+=eval('dt')
    
    def Reset(self):
        '''Setzt den Wert auf den Startwert zurueck.'''
        self._Faktor__wert=self.GibStartwert()
        self._Faktor__alleWerte=[self.GibStartwert()]

#### Simulation ----------------------------------------------------------------
#### die Klasse Simulation initialisiert und steuert die Simulation ============
class Simulation():
    '''Simulation stellt die zusammenfassende Klasse für die Modellbeschreibung dar.
    Sie uebernimmt die Funktion eines Controllers.'''

    die_Simulation = None

    @staticmethod
    def GibSimulation(view=None):
        '''Singleton ermoeglichen'''
        if Simulation.die_Simulation == None:
            Simulation.die_Simulation = Simulation()
            Simulation.die_Simulation.SetzeView(view)
        return Simulation.die_Simulation

    def __init__(self):
        '''Die Listen enthalten die Namen der Faktoren'''
        self.__modellName=None
        self.__modellDaten=[]
        self.__bestandsfaktoren=[]
        self.__fluesse=[]
        self.__parameter=[]
        self.__zwischengroessen=[]
        # Wirkungen werden in der Klasse Wirkungen gehalten
        self.__t=Zeit.GibZeit()
        self.__anzahl_Schritte=10
        self.LoescheWerte()
        self.__simulationsMethode='erweitert Euler'
        ## zulaessig: 'einfach Euler','erweitert Euler','Runge-Kutta'
        self.__ObjekteGeloescht=[]

    def SetzeModellNamen(self, name):
        assert name.__class__==str, "String eingeben!"
        self.__modellName=name

    def GibModellNamen(self):
        return self.__modellName

    def AendereFaktorNamen(self, alterName, neuerName):
        if not(self.NeuerNameOK(neuerName)):
            self.Meldung("Name existiert bereits!")
            return False
        self.ModellWerteZuruecksetzen()
        faktor=eval(alterName)
        if faktor.__class__==Bestandsfaktor:
            self.__bestandsfaktoren.remove(faktor.GibNamen())
            faktor.SetzeNeuenNamen(neuerName)
            self.__bestandsfaktoren.append(neuerName)            
        elif faktor.__class__==Fluss:
            self.__fluesse.remove(faktor.GibNamen())
            for bestandsName in self.__bestandsfaktoren:
                bestand=eval(bestandsName)
                bestand.AendereFlussNamen(alterName,neuerName)
            faktor.SetzeNeuenNamen(neuerName)
            self.__fluesse.append(neuerName)
        elif faktor.__class__==Parameter:
            self.__parameter.remove(faktor.GibNamen())
            faktor.SetzeNeuenNamen(neuerName)
            self.__parameter.append(neuerName)
        elif faktor.__class__==Zwischengroesse:
            self.__zwischengroessen.remove(faktor.GibNamen())
            faktor.SetzeNeuenNamen(neuerName)
            self.__zwischengroessen.append(neuerName)
        ## Alle informieren, die einen Term haben:
        for faktorName in self.__fluesse+self.__zwischengroessen:
            faktor=Faktor.FaktorZuNamen(faktorName)
            faktor.UpdateTerm(alterName, neuerName)
        ## Fluesse informieren ueber Aenderung fuer aus / in
        for flussName in self.__fluesse:
            fluss=Faktor.FaktorZuNamen(flussName)
            fluss.AendereAusgangsBestand(alterName, neuerName)
            fluss.AendereZielBestand(alterName, neuerName)
        ## Wirkungen informieren:
        for wirkung in Wirkung.GibAlleWirkungen():
            wirkung.FaktorNameGeaendert(alterName,neuerName)
        return True

    def GibAlleFaktorNamen(self):
        ergebnis=[]
        ergebnis+=self.__bestandsfaktoren
        ergebnis+=self.__fluesse
        ergebnis+=self.__parameter
        ergebnis+=self.__zwischengroessen
        return ergebnis

    def GibFaktorNamen(self, faktorTyp=None):
        if faktorTyp==None:
            return self.GibAlleFaktorNamen()
        if faktorTyp=='Bestandsfaktor':
            return self.__bestandsfaktoren
        if faktorTyp=='Fluss':
            return self.__fluesse
        if faktorTyp=='Parameter':
            return self.__parameter
        if faktorTyp=='Zwischengroesse':
            return self.__zwischengroessen
         # sonst alle anbieten        
        else:
            return []

    def NeuerNameOK(self, name):
        if name in self.__bestandsfaktoren: return False
        if name in self.__fluesse: return False
        if name in self.__parameter: return False
        if name in self.__zwischengroessen: return False
        if name in globals().keys(): return False
        return True

    def NeuerBestandsfaktor(self, name, startwert=None):
        '''Zu allen Faktoren werden nur die Namen gespeichert
        und der Faktor ist auch nur mit seinem Namen angegeben!'''
        if not self.NeuerNameOK(name):
            self.Meldung("Name "+name+" existiert bereits.")
            return None
        self.__bestandsfaktoren.append(name)
        if startwert==None: return Bestandsfaktor(name)
        else: return Bestandsfaktor(name,startwert)
        
    def NeuerFluss(self, name, aus_Bestand, in_Bestand):
        '''Zu allen Faktoren werden nur die Namen gespeichert
        und der Faktor ist auch nur mit seinem Namen angegeben!'''
        if not self.NeuerNameOK(name):
            self.Meldung("Name "+name+" existiert bereits.")
            return None
        self.__fluesse.append(name)
        return Fluss(name, aus_Bestand, in_Bestand)
        
    def NeuerParameter(self, name, startwert=None):
        '''Zu allen Faktoren werden nur die Namen gespeichert
        und der Faktor ist auch nur mit seinem Namen angegeben!'''
        if not self.NeuerNameOK(name):
            self.Meldung("Name "+name+" existiert bereits.")
            return None
        self.__parameter.append(name)
        if startwert==None: return Parameter(name)
        else: return Parameter(name,startwert)
        
    def NeueZwischengroesse(self, name):
        '''Zu allen Faktoren werden nur die Namen gespeichert
        und der Faktor ist auch nur mit seinem Namen angegeben!'''
        if not self.NeuerNameOK(name):
            self.Meldung("Name "+name+" existiert bereits.")
            return None
        self.__zwischengroessen.append(name)
        return Zwischengroesse(name)

    def SetzeTerm(self, faktorName, term):
        faktor = Faktor.FaktorZuNamen(faktorName)
        return faktor.SetzeTerm(term)
        
    def NeueWirkung(self, vonName, aufName):
        '''Wirkungen werden nicht mit Namen gespeichert sondern nur
        als Objekte in der Liste.'''
        if vonName==aufName: return None # keine wirkung auf sich selbst
        wirkung=Wirkung(vonName, aufName)
        if wirkung.undefiniert: return None
        return wirkung

    def LoescheBestandsfaktor(self,name):
        '''entfernt das Bestandsfaktorobjekt zum Namen'''
        bestand=Faktor.FaktorZuNamen(name)
        if bestand!=None:
            ## Alle Wirkungen und alle Fluesse loeschen, die diesen
            ## Bestandsfaktor betreffen.
            for wirkung in self.GibWirkungen():
                if (name==wirkung.GibVon()) or (name==wirkung.GibAuf()):
                    self.LoescheWirkung(wirkung.GibVon(),wirkung.GibAuf())
            for flussName in bestand.GibZuFluesse():
                self.LoescheFluss(flussName)
            for flussName in bestand.GibAbFluesse():
                self.LoescheFluss(flussName)
            Faktor.LoescheFaktor(name)
            self.__bestandsfaktoren.remove(name)
            self.__ObjekteGeloescht.append(name)
        else:
            self.Meldung("Bestandsfaktor "+name+" existiert nicht!")
        
    def LoescheFluss(self,name):
        '''entfernt das Flussobjekt zum Namen'''
        fluss=Faktor.FaktorZuNamen(name)
        if fluss!=None:
            ## Alle Wirkungen loeschen, die diesen Fluss betreffen.
            for wirkung in self.GibWirkungen():
                if name in (wirkung.GibVon(),wirkung.GibAuf()):
                    self.LoescheWirkung(wirkung.GibVon(),wirkung.GibAuf())
            ## Alle Bestandsfaktoren informieren
            for bestandsName in self.__bestandsfaktoren:
                bestand=Faktor.FaktorZuNamen(bestandsName)
                if bestand!=None:
                    bestand.Entferne_Fluss(name)
            Faktor.LoescheFaktor(name)
            self.__fluesse.remove(name)
            self.__ObjekteGeloescht.append(name)
        else:
            self.Meldung("Fluss "+name+" existiert nicht!")
    
    def LoescheParameter(self,name):
        '''entfernt den Parameter zum Namen'''
        parameter=Faktor.FaktorZuNamen(name)
        if parameter!=None:
            ## Alle Wirkungen loeschen, die diesen Parameter betreffen.
            for wirkung in self.GibWirkungen():
                if name in (wirkung.GibVon(),wirkung.GibAuf()):
                    self.LoescheWirkung(wirkung.GibVon(),wirkung.GibAuf())
            Faktor.LoescheFaktor(name)
            self.__parameter.remove(name)
            self.__ObjekteGeloescht.append(name)
        else:
            self.Meldung("Parameter "+name+" existiert nicht!")
    
    def LoescheZwischengroesse(self,name):
        '''entfernt die Zwischengroesse zum Namen'''
        zg=Faktor.FaktorZuNamen(name)
        if zg!=None:
            ## Alle Wirkungen loeschen, die diese Zwischengroesse betreffen.
            for wirkung in []+self.GibWirkungen(): # echte KOpie erzeugen
                if name in wirkung.GibVon()+wirkung.GibAuf():
                    self.LoescheWirkung(wirkung.GibVon(),wirkung.GibAuf())
            Faktor.LoescheFaktor(name)
            self.__zwischengroessen.remove(name)
            self.__ObjekteGeloescht.append(name)
        else:
            self.Meldung("Zwischengroesse "+name+" existiert nicht!")
    
    def LoescheWirkung(self, vonName, aufName):
        wirkung=Wirkung.HatWirkung(vonName, aufName)
        if not(wirkung): return False # existiert nicht
        self.__ObjekteGeloescht.append(wirkung.GibVon()+'->'+wirkung.GibAuf())
        Wirkung.LoescheWirkung(wirkung)

    def HoleGeloeschte(self):
        geloescht=[]+self.__ObjekteGeloescht # echte Kopie
        self.__ObjekteGeloescht=[]
        return geloescht

    def GibWirkungen(self):
        return Wirkung.GibAlleWirkungen()

    def SetzeSimulationsMethode(self, methode='erweitert Euler'):
        if methode in ['erweitert Euler', 'einfach Euler','Runge-Kutta']:
            self.__simulationsMethode=methode
    
    #### ------- Iterationsschritt berechnen ----------------------------------
    def BerechneWerte(self, groessen):
        if self.__simulationsMethode=='einfach Euler':
            self.BerechneWerteEinfach(groessen)
        elif self.__simulationsMethode=='erweitert Euler':
            self.BerechneWerteErweitert(groessen)
        else:
            self.BerechneWerteRungeKutta(groessen)

    def BerechneWerteEinfach(self, groessen):# anzeigeBestand, zweiter=None):
        # initialisieren
        self.LoescheWerte()
        bestaende = map(lambda x:eval(x,globals()),self.__bestandsfaktoren)
        ## Versionsunabhaengigkeit von map herstellen:
        if versionsNummer=='3': bestaende=list(bestaende)
        self.SetzeAusgabeWerte(self.AktuelleWerteHolen(groessen))
        # Iteration
        for i in range(self.__anzahl_Schritte):
            for bestand in bestaende:
                bestand.BerechneNeuenWert()
            ## erst alle Werte neu berechnen und dann erst neu setzen
            for bestand in bestaende:
                bestand.UebernimmNeuenWert()
            self.__t.Schritt()
            self.SetzeAusgabeWerte(self.AktuelleWerteHolen(groessen))
            self.SetzeAlleWerte()

    def BerechneWerteErweitert(self, groessen):
        """Simulationsmethode:'erweitert Euler'.
        Ist der erste Parameter der groessen 't', wird ein Zeitdiagramm,
        sonst ein Phasendiagramm gezeichnet"""
        # initialisieren
        self.LoescheWerte()
        bestaende = map(lambda x:eval(x,globals()),self.__bestandsfaktoren)
        ## Versionsunabhaengigkeit von map herstellen:
        if versionsNummer=='3': bestaende=list(bestaende)
        self.SetzeAusgabeWerte(self.AktuelleWerteHolen(groessen))
        for i in range(self.__anzahl_Schritte):
            self.__zeitKorrektur = 0
            # Werte am Anfang des Intervalls sichern
            bestandswerte_0dt = []
            bestandswerte_2dt = []
            for bestand in bestaende:
                bestandswerte_0dt.append(bestand.GibWert())
                bestand.BerechneNeuenWert()
            for bestand in bestaende:
                bestand.UebernimmNeuenWert()
            for bestand in bestaende:
                bestand.BerechneNeuenWert()
            for bestand in bestaende:
                bestand.UebernimmNeuenWert()
                bestandswerte_2dt.append(bestand.GibWert())
            j = -1
            for bestand in bestaende:
                j += 1
                bestand.UebernimmNeuenWert((bestandswerte_0dt[j] + bestandswerte_2dt[j]) / 2.0)
            self.__t.Schritt()
            self.SetzeAusgabeWerte(self.AktuelleWerteHolen(groessen))
            self.SetzeAlleWerte()

    def BerechneWerteRungeKutta(self, groessen):
        """Simulationsmethode:'Runge-Kutta'.
        Ist der erste Parameter der groessen 't', wird ein Zeitdiagramm,
        sonst ein Phasendiagramm gezeichnet"""
        # initialisieren
        self.LoescheWerte()
        bestaende = map(lambda x:eval(x,globals()),self.__bestandsfaktoren)
        ## Versionsunabhaengigkeit von map herstellen:
        if versionsNummer=='3': bestaende=list(bestaende)
        self.SetzeAusgabeWerte(self.AktuelleWerteHolen(groessen))
        for i in range(self.__anzahl_Schritte):            
            b_0,b_1,b_2,b_3,b_4,b_5,b_6 = [],[],[],[],[],[],[]
            # wegen der Verstaendlichkeit wird mit m1,m2,m3,m4 gearbeitet
            m1,m2,m3,m4 = [],[],[],[]
            for bestand in bestaende:
                # Werte am Anfang des Intervalls
                b_0.append(bestand.GibWert())
                bestand.BerechneNeuenWert(halberSchritt=True) # B1
            for bestand in bestaende:
                bestand.UebernimmNeuenWert() # Stand B1
                b_1.append(bestand.GibWert())
            for i in range(len(bestaende)):
                m1.append((b_1[i]-b_0[i])*2.0/dt)
            for bestand in bestaende:
                bestand.BerechneNeuenWert(halberSchritt=True) # B2
            for bestand in bestaende:
                bestand.UebernimmNeuenWert() # Stand B2
                b_2.append(bestand.GibWert())
            for i in range(len(bestaende)):
                m2.append((b_2[i]-b_1[i])*2.0/dt)
                # Zustand B3 (bei t+dt/2) erzeugen: B0+m2*dt/2
                bestaende[i].UebernimmNeuenWert(b_0[i]+m2[i]*dt/2.0) # Stand B3
                b_3.append(bestaende[i].GibWert())
            for bestand in bestaende:
                bestand.BerechneNeuenWert(halberSchritt=True) # B4
            for bestand in bestaende:
                bestand.UebernimmNeuenWert() # Stand B4
                b_4.append(bestand.GibWert())
            for i in range(len(bestaende)):
                m3.append((b_4[i]-b_3[i])*2.0/dt)
                # Zustand B5 (bei t+dt) erzeugen: B0+m3*dt
                bestaende[i].UebernimmNeuenWert(b_0[i]+m3[i]*dt) # Stand B5
                b_5.append(bestaende[i].GibWert())
            for bestand in bestaende:
                bestand.BerechneNeuenWert() # B6
            for bestand in bestaende:
                bestand.UebernimmNeuenWert() # Stand B6
                b_6.append(bestand.GibWert())
            for i in range(len(bestaende)):
                m4.append((b_6[i]-b_5[i])/dt)
                # Iterationswert bestimmen
                wert=b_0[i]+(m1[i]+2*m2[i]+2*m3[i]+m4[i])*dt/6.0
                # Formelvariante ohne m1,m2,m3,m4:
                #wert=b_0[i]+(2*(b_1[i]-b_0[i])+4*(b_2[i]-b_1[i])+4*(b_4[i]-b_3[i])+b_6[i]-b_5[i])*dt/6.0
                bestaende[i].UebernimmNeuenWert(wert)
            ##print b_0,b_1,b_2,b_3,b_4,b_5,b_6
            ##print m1,m2,m3,m4,'\n'
            self.__t.Schritt()
            self.SetzeAusgabeWerte(self.AktuelleWerteHolen(groessen))
            self.SetzeAlleWerte()

    def AktuelleWerteHolen(self, faktorNamenListe):
        ergebnis=[]
        for faktorName in faktorNamenListe:
            faktor=eval(faktorName)
            ergebnis.append(faktor.GibWert()) # hier sollte keine Verzoegerung wirken
        return tuple(ergebnis)

    def SetzeAlleWerte(self):
        for faktor in Faktor.alleFaktorNamen:
            eval(faktor).UpdateAlleWerte()

    def SetzeAusgabeWerte(self, werte):
        '''fuegt die auszugebenden Werte des aktuellen Schritts hinzu'''
        self.__werte.append(werte)

    def GibWerte(self):
        return self.__werte

    def LoescheWerte(self):
        self.__werte = []
        self.ModellWerteZuruecksetzen()
        
    def Setze_dt(self, dt=1.0):
        '''setzte den globalen Wert dt durch das Zeit-Objekt neu'''
        self.__t.Setze_dt(dt)

    def Gib_dt(self):
        '''greift auf die globale Variable dt zu'''
        return eval(dt)

    def SetzeAnzahlSchritte(self, anzahl):
        '''Setzt die Anzahl der Schritte'''
        self.__anzahl_Schritte=anzahl
    
#### Abschnitt für die Modelldefinition -----------------------------
    def GibModellDaten(self):
        return self.__modellDaten

    def ZeigeModellDaten(self):
        '''Modelldaten zeilenweise lesbar'''
        daten=self.__modellDaten
        zeilen=['Bestandsfaktoren:']
        if (len(daten)>0) and (daten[0][0]!='Bestand'):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Bestand'):
            zeilen.append(daten[0][1]+' = '+str(daten[0][2]))
            daten=daten[1:]
        zeilen.append('\nFluesse:')
        if (len(daten)>0) and (daten[0][0]!='Fluss'):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Fluss'):
            von,nach=daten[0][2],daten[0][3]
            ##if von=='': von='Quelle'
            ##if nach=='': nach='Senke'
            zeilen.append(daten[0][1]+': '+von+' => '+nach)
            daten=daten[1:]
        zeilen.append('\nParameter')
        if (len(daten)>0) and (daten[0][0]!='Parameter'):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Parameter'):
            zeilen.append(daten[0][1]+': '+str(daten[0][2]))
            daten=daten[1:]
        zeilen.append('\nZwischengroesse')
        if (len(daten)>0) and (daten[0][0]!='Zwischengroesse'):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Zwischengroesse'):
            zeilen.append(daten[0][1])
            daten=daten[1:]            
        zeilen.append('\nWirkungen:')
        if (len(daten)>0) and (daten[0][0]!='Wirkung'):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Wirkung'):
            zeilen.append(daten[0][1]+' --> '+daten[0][2])
            daten=daten[1:]
        zeilen.append('\nTerme:')
        if (len(daten)>0) and (daten[0][0]!='Term'):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Term'):
            zeilen.append(daten[0][1]+'='+daten[0][2])
            daten=daten[1:]
        zeilen.append('\nTabellendaten:')
        if (len(daten)==0) or ((len(daten)>0) and (daten[0][0]!='Tabellendaten')):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Tabellendaten'):
            zeilen.append(daten[0][1]+'='+str(daten[0][2]))
            daten=daten[1:]
        return zeilen
    
    def SichereModellDaten(self):
        self.__modellDaten=[]
        if self.__modellName==None:
            self.SetzeModellNamen('Neu')
        for bestandsName in self.__bestandsfaktoren:
            bestand=Faktor.FaktorZuNamen(bestandsName)
            self.__modellDaten.append(
                ('Bestand',bestandsName,
                 bestand.GibStartwert()))
        for flussName in self.__fluesse:
            fluss=Faktor.FaktorZuNamen(flussName)
            self.__modellDaten.append(
                ('Fluss',
                 flussName,
                 fluss.GibAusgangsBestand(),
                 fluss.GibZielBestand()))
        for parameterName in self.__parameter:
            parameter=Faktor.FaktorZuNamen(parameterName)
            self.__modellDaten.append(
                ('Parameter',parameterName,parameter.GibStartwert()))
        for name in self.__zwischengroessen:
            self.__modellDaten.append(('Zwischengroesse',name))
        for wirkung in Wirkung.GibAlleWirkungen():
            self.__modellDaten.append(
                ('Wirkung',wirkung.GibVon(),wirkung.GibAuf()))
        ## nur Fluesse und Zwischengroessen haben Terme mit Faktoren
        for faktorName in self.__fluesse+self.__zwischengroessen:
            faktor=Faktor.FaktorZuNamen(faktorName)
            self.__modellDaten.append(
                ('Term',faktorName,faktor.GibTerm()))        
        ## nur Parameter haben ggf Terme ohne Faktoren
        for faktorName in self.__parameter:
            faktor=Faktor.FaktorZuNamen(faktorName)
            if faktor.GibTerm()!=None:
                self.__modellDaten.append(('Term',faktorName,faktor.GibTerm()))
        ## nur Parameter haben Tabellenwerte oder Terme ohne Faktoren
        for faktorName in self.__parameter:
            faktor=Faktor.FaktorZuNamen(faktorName)
            if faktor.GibTabelle()!=None:
                self.__modellDaten.append(('Tabellendaten',faktorName,faktor.GibTabellenWerte()))        
        
    def ErzeugeModellAusDaten(self, modell_Daten):
        for daten in modell_Daten:
            if daten[0]=='Bestand':
                if len(daten)==2:
                    self.NeuerBestandsfaktor(daten[1])
                else:
                    self.NeuerBestandsfaktor(daten[1],daten[2])
            elif daten[0]=='Fluss':
                if len(daten)==2:
                    self.Meldung('Modellfehler bei '+daten[1])
                    return
                elif len(daten)==2:
                    self.NeuerFluss(daten[1],daten[2])
                else:
                    self.NeuerFluss(daten[1],daten[2],daten[3])
            elif daten[0]=='Wirkung':
                self.NeueWirkung(daten[1],daten[2])
            elif daten[0]=='Parameter':
                self.NeuerParameter(daten[1],daten[2])
            elif daten[0]=='Zwischengroesse':
                self.NeueZwischengroesse(daten[1])
            elif daten[0]=='Term':
                self.SetzeTerm(daten[1],daten[2])
            elif daten[0]=='Tabellendaten':
                parameter=Faktor.FaktorZuNamen(daten[1])
                zeiten,werte=daten[2]
                parameter.SetzeTabelle(zeiten,werte)
            elif daten[0]=='dt':
                self.Setze_dt(daten[1])
            elif daten[0]=='AnzahlSchritte':
                self.SetzeAnzahlSchritte(daten[1])
            else:
                self.Meldung('Fehler bei Modelldefinition: '+daten[0])

    def ModellNameAusDateiNamen(self, dateiname):
        # Pfad abspalten
        if '\\' in dateiname: # windows-Pfadangabe
            liste=dateiname.rsplit('\\')
        else:
            liste=dateiname.rsplit('/')
        liste.reverse()
        name=liste[0]
        if '.' in name:
            name=name.rsplit('.')[0]
        return name

    def ModellWerteZuruecksetzen(self):
        for name in Faktor.GibAlleFaktorNamen():
            eval(name).Reset()
        self.__t.Reset()

    def LoescheAktuellesModell(self):
        self.__modellName=None
        self.__modellDaten=[]
        wirkungen=[]+Wirkung.GibAlleWirkungen()
        for wirkung in wirkungen: Wirkung.LoescheWirkung(wirkung)
        fluesse=[]+self.__fluesse # echte Kopie notwendig!
        for fluss in fluesse: self.LoescheFluss(fluss)
        faktoren=[]+self.__bestandsfaktoren
        for faktor in faktoren: self.LoescheBestandsfaktor(faktor)
        parameter=[]+self.__parameter
        for par in parameter: self.LoescheParameter(par)
        zwischengroessen=[]+self.__zwischengroessen
        for zwi in zwischengroessen: self.LoescheZwischengroesse(zwi)
        
#### Abschnitt für das Laden und Speichern von Modelldaten -----------------
    def Speichere(self, dateiname='', modellDaten=[]):
        try:
            self.SichereModellDaten()
            if dateiname=='':
                if self.__modellName==None: return None
                dateiname=self.__modellName+'.simdat'
            f=open(dateiname,'w')
            if modellDaten!=[]: f.write(str(modellDaten))
            else: f.write(str(self.__modellDaten))
            f.close()
        except IOError as e:
            self.Meldung('Fehler beim Schreiben: '+str(e))
            return None

    def Lies(self, dateiname):
        try:
            f=open(dateiname,'r')
            modell_Daten_String=f.read()
            f.close()
            self.__modellName=self.ModellNameAusDateiNamen(dateiname)
            self.__modell_Daten=eval(modell_Daten_String)
            self.__modellName
            self.ErzeugeModellAusDaten(self.__modell_Daten)
            return self.__modell_Daten
        except IOError as e:
            self.Meldung('Fehler beim Lesen: '+str(e))
            return None

#### Abschnitt für die Simulationsdurchführung -----------------------------
    def Starte_Simulation(self, groessen):
        self.LoescheWerte()
        self.__zeitKorrektur = 0
        ##self.ModellWerteZuruecksetzen() passiert in LoescheWerte()
        self.BerechneWerte(groessen)
        return self.GibWerte()
        
    def ZeigeHilfe(self, n=1):
        if n==1:
            hinweise=(u"Bestandsfaktoren mit Namen und Anfangswert definieren.\n"
                      +u"  (Begriff Faktoren entspricht Größen)\n"
                      +u"Der Anfangswert kann auch nachträglich gesetzt werden.\n\n"
                      +u"Fluesse muessen bei ihrer Definition\n"
                      +u"mindestens einen Bestandsfaktor\n"
                      +u" -> als Ausgangsbestand (aus_Bestand) oder\n"
                      +u" -> als Zielbestand (in_Bestand) haben.\n\n"
                      +u"=> Bestandsfaktoren vor den Fluessen definieren.\n\n"
                      +u"Faktornamen Python - konform, also kein Minuszeichen, "
                      +u"kein Leerzeichen, keine Sonderzeichen;\n"
                      +u"Faktornamen in Termen mit () beenden,"
                      +u" damit das Auswerten ausgelöst wird.\n"
                      +u"Eine Verzoegerung kann in der Klammer eingegeben"
                      +u"werden, beispielsweise: Fische(3) ."
                      )
        elif n==2:
            hinweise=("Muster für die Modelldefinition:\n"
                      +"Dateiname: '<Modellname>.simdat (s.u.!)'\n\n"
                      +"Inhalte sind Listen von Tupeln:\n"
                      +" ('Bestand',<Name>, <Anfangswert>)\n oder"
                      +" ('Bestand',<Name>)\n"
                      +" ('Fluss',<Name>, <aus_Bestand>/None, <in_Bestand>/None)\n"
                      +" ('Parameter',<Name>, <Wert>)\n"
                      +" ('Zwischengroesse', <Name>)\n"
                      +" ('Wirkung',<von_Faktor>, <auf_Faktor>)\n"
                      +" ('Term',<Faktor_Name,<Term>)\n"
                      +"   Faktornamen im Term mit () beenden,"
                      +" damit das Auswerten ausgelöst wird.\n"
                      +" Verzoegerung in der Klammer eingeben,"
                      +"  beispielsweise: Fische(3) .\n\n"
                      +"In Dateien mit der Endung simdia sollen die Tupel\n"
                      +"   von Bestand, Parameter oder Zwischengroesse\n"
                      +" jeweils am Ende die Position im Modellfenster als Tupel enthalten.")
        return hinweise
    
    def SetzeView(self, view):
        self.__view=view
        
    def Meldung(self, text):
        if self.__view==None: print(text)
        else: self.__view.Meldung(text)

#### Tests in der Datei simu_test.py
if __name__=='__main__':
    simu=Simulation.GibSimulation()
##    simu.Lies('Modelle/Bakterienwachstum-mit-altemWert.simdia')
##    simu.BerechneWerteErweitert(['t','Bakterien'])
##    ## Loeschen testen
##    simu.LoescheAktuellesModell()
##    simu.Lies('Modelle/Hasen&Fuechse.simdia')
##    simu.BerechneWerteErweitert(['t','Hasen','Fuechse'])
##    ## Loeschen testen
##    simu.LoescheAktuellesModell()
##    simu.Lies('Modelle/Neubritannien.simdia')
##    simu.BerechneWerteErweitert(['Gaeste','Umwelt'])
##    simu.Meldung(simu.GibWerte())
##    simu.SichereModellDaten()
##    simu.Meldung(simu.ZeigeModellDaten())
##    print('\n\n')
##    simu.LoescheAktuellesModell()
##    simu.SichereModellDaten()
##    simu.Meldung(simu.ZeigeModellDaten())
##    print('\n\n')
##    simu.Lies('Modelle/Neubritannien.simdia')
##    simu.SichereModellDaten()
##    simu.Meldung(simu.ZeigeModellDaten())

##    ## Fehler beim Schreiben testen
##    simu.Speichere('Modelle/Modelle/son_Mist.simdia')
##    ## Fehler beim Lesen testen
##    simu.Lies('Modelle/son_Mist.simdia')
    

